//
//  decode.c
//  bpic
//
//  Created by theBreadCompany on 04.03.23.
//

#include "decode.h"

void read_to_buffer(png_structp png_ptr, png_bytep data, png_size_t length);

decoded_data *bpic_decode_memory(encoded_data *data, bool *checksum_matching) {
    
    if(!data->buffer) { // is there even data
        throw_err(bpic_error_input_buffer_null, 0);
        return (void*)bpic_error_input_buffer_null;
    }
    if(data->size < 7) { // is there enough data to hold at least a header
        throw_err(bpic_error_input_buffer_small, 0);
        return (void*)bpic_error_input_buffer_small;
    }
    
    png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    png_infop png_info = png_create_info_struct(png_ptr);
    
    //printf("%zu\n", data->size);
    /*
    mem_decode *input = malloc(sizeof(mem_decode));
    input->buffer = data->buffer;
    input->size = data->size;
    input->pos = 0;
    png_set_read_fn(png_ptr, input, read_to_buffer);
    */

    decoded_data *output = bpic_decode_file(fmemopen(data->buffer, data->size, "r"), checksum_matching);
    
    png_free(png_ptr, png_info);
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    
    //free(png_ptr);
    return output;
}

decoded_data* bpic_decode_file(FILE *input, bool *checksum_matching) {
    fseek(input, 0, SEEK_SET);
    
    png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    png_infop png_info = png_create_info_struct(png_ptr);
    
    if(input) {
        png_init_io(png_ptr, input);
    } else {
        printf("Failed to init png with file!\n");
        return NULL;
    }
    
    image_data *output = bpic_decode_png(png_ptr, png_info, checksum_matching);

    png_free(png_ptr, png_info);
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    
    return output;
}

decoded_data* bpic_decode_png(png_structp png_ptr, png_infop png_info, bool *checksum_matching) {
    
    png_set_keep_unknown_chunks(png_ptr, PNG_HANDLE_CHUNK_IF_SAFE, (png_bytep)"pvSF", 1);
    
    png_read_info(png_ptr, png_info);
        
    printf("%u x %u\n", png_get_image_height(png_ptr, png_info), png_get_image_width(png_ptr, png_info));
    
    png_unknown_chunkp chunks;
    int chunk_count = png_get_unknown_chunks(png_ptr, png_info, &chunks);

    bpic_info bpic_chunk;
    bpic_chunk.length = 0;
    if(chunk_count) { // are there unknown chunks
        for(uint8_t i = 0; i < chunk_count; i++) { // for every unknown chunk
            png_unknown_chunk chunk = chunks[i]; // store chunk
            
            char *name = malloc(5); // alloc name
            memcpy(name, chunk.name, 5); // copy name (same procedure as setting it
            if(!strcmp(name, "pvSF") && chunk.size == 7) { // if name is pvSF header name and size is 12 bytes
                bpic_chunk.px_size = pow(2, chunk.data[0] >> 3); // remember, the stored pxsize is a power of 2 leftshifted by 2
                bpic_chunk.encoding = chunk.data[0] & 0b11;
                memcpy(&bpic_chunk.length, chunk.data + 1, 4);
                memcpy(&bpic_chunk.fletcher16, chunk.data + 5, 2);
                
                for(int byte = 0; byte < 7; byte++) {
                    for(int bit = 1; bit < 9; bit++) {
                        printf("%s", chunk.data[byte] >> (8 - bit) & 1 ? "1" : "0");
                    }
                    printf(" ");
                }
                printf("\n");
            }
            free(name);
            free(chunk.data);
        }
    } else {
        throw_err("input is missing unknown chunks, meaning that the pvSF chunk is also missing!", 0);
        return (void*)"input is missing unknown chunks, meaning that the pvSF chunk is also missing!";
    }
    
    if(!bpic_chunk.length) {
        fprintf(stderr, "Missing data length in bpic chunk!\n");
        return NULL;
    }

    free(chunks);
    printf("pxsize: %u; encoding: %u; length: %zu; checksum: %u\n", bpic_chunk.px_size, bpic_chunk.encoding, bpic_chunk.length, bpic_chunk.fletcher16);
    
    int width = png_get_image_width(png_ptr, png_info);// * bpic_chunk.px_size;
    int height = png_get_image_height(png_ptr, png_info);// * bpic_chunk.px_size;
    
    png_bytep *row_data = png_malloc(png_ptr, sizeof(png_bytep) * height * bpic_chunk.px_size);
    for(int y = 0; y < height; y++) row_data[y] = png_malloc(png_ptr, sizeof(png_byte) * width * bpic_chunk.px_size);
    png_set_rows(png_ptr, png_info, row_data);
    png_read_image(png_ptr, row_data);
    png_read_end(png_ptr, NULL);

    
    int bits = png_get_bit_depth(png_ptr, png_info);
    
    // create a filebuffer that can hold as many bytes as needed according to the header
    unsigned char *filebuffer = malloc(bpic_chunk.length);//(width/bits * height/bits);//(bpic_chunk.length);
    // if encoding is bitwise
    if(!bpic_chunk.encoding) {
        // if the data is encoded in an 8 bit grayscale image
        if(bits == 8) {
            // for every [px_size]th in the data; also init a bit count
            for(int y = 0, bit_i = 0; y < height; y += bpic_chunk.px_size) {
                // for every [px_size]th column; also increment the bit count
                for(int x = 0; x < width && (bit_i >> 3) < bpic_chunk.length; x += bpic_chunk.px_size, bit_i++) {
                    // store the pixel data
                    png_byte px = row_data[y][x * bpic_chunk.px_size];
                    // store the corresponding the data
                    uint8_t value = px & (0x80 >> (bit_i & 0b111));
                    // add the data in the byte of the buffer; 8 bit = 1 byte, so divide the bit count by 8 to get the byte position
                    filebuffer[bit_i >> 3] += value;
                }
            }
        // if the data is encoded as a 1 bit grayscale image
        } else if(bits == 1) {
            // the image widths are required to be divisible by 8 to make parsing easier
            // basically same algorithm, but now 
            for(int y = 0, bit_i = 0; y < height; y += bpic_chunk.px_size) {
                // for every [px_size]th column; also increment the bit count
                for(int x = 0; x < width && (bit_i >> 3) < bpic_chunk.length; x += bpic_chunk.px_size, bit_i++) {
                    // store the pixel data
                    png_byte px = row_data[y][x * bpic_chunk.px_size];
                    // store the corresponding the data
                    uint8_t value = px & (0x80 >> (bit_i & 0b111));
                    // add the data in the byte of the buffer; 8 bit = 1 byte, so divide the bit count by 8 to get the byte position
                    filebuffer[bit_i >> 3] += value;
                }
            }
        // unknown bit count, cancel...
        } else {
            
        }
    // othrwise the encoding is bytewise
    } else {
        // this is far more straightforward as the rows literally represent the data
        if(bpic_chunk.px_size == 1) {
            // every row, except the last as it may break the available range of the file buffer
            for(int y = 0; y < height - 1; y++) {
                // copy the row to the filebuffer with an offset of the number of already processed bytes
                memcpy(filebuffer + y * width, row_data[y], width);
            }
            // the required data size to be copied is calculated as the required length minus the already parsed length
            memcpy(filebuffer + (height - 1) * width, row_data[height-1], bpic_chunk.length - width * (height - 1));
        // except if the pixel size is > 1, then only every [px_size]th pixel is valid data
        } else {
            // for every [px_size]th row except the last
            for(int y = 0; y < height - bpic_chunk.px_size; y += bpic_chunk.px_size) {
                // for every [px_size]th column
                for(int x = 0; x < width; x += bpic_chunk.px_size) {
                    // copy every [px_size]th byte to the buffer at offset of the current byte
                    memcpy(filebuffer + y / bpic_chunk.px_size * width + x / bpic_chunk.px_size, row_data[y] + x * bpic_chunk.px_size, 1);
                }
            }
            // separatly go through the last row as it may contain more data than required
            for(int x = 0; x < bpic_chunk.length - (height - 1) * width; x += bpic_chunk.px_size) {
                memcpy(filebuffer + (height - 1) * width + x / bpic_chunk.px_size, row_data[height - 1] + x * bpic_chunk.px_size, 1);
            }
        }
    }
    for(int y = 0; y < height; y++) png_free(png_ptr, row_data[y]);
    png_free(png_ptr, row_data);
    
    uint16_t checksum = fletcher16(filebuffer, bpic_chunk.length);
    printf("Checksum of decoded data: %u\n", checksum);
    *checksum_matching = bpic_chunk.fletcher16 == checksum;
    
    
    decoded_data *output = malloc(sizeof(decoded_data));
    output->buffer = filebuffer;
    output->size = bpic_chunk.length;
    
    return output;
}

void read_to_buffer(png_structp png_ptr, png_bytep data, png_size_t length) {
    mem_decode *data_info = png_get_io_ptr(png_ptr);
    if(data_info->size < data_info->pos + length) {
        printf("How... The input is bigger than expected, already at %lu/%zu\n", data_info->pos + length, data_info->size);
    }
    memcpy(data, data_info->buffer + data_info->pos, length);
    data_info->pos += length;
}
