//
//  encode.c
//  bpic
//
//  Created by theBreadCompany on 04.03.23.
//

#include "encode.h"

void write_to_buffer(png_structp png_ptr, png_bytep data, png_size_t length);
                
png_bytepp convert_bitwise_1bit_grayscale(image_data *data, png_structp png_ptr, png_infop info_ptr);
png_bytepp convert_bitwise_8bit_grayscale(image_data *data, png_structp png_ptr, png_infop info_ptr);
png_bytepp convert_bytewise_8bit_rgb(image_data *data, png_structp png_ptr, png_infop info_ptr);

png_bytepp convert_bitwise_1bit_grayscale(image_data *data, png_structp png_ptr, png_infop info_ptr) {
    int width = round(sqrt((double)(data->size << 3)));
    int height = round((data->size << 3) / (double)width);
    
    png_bytepp row_data = convert_bytewise_8bit_rgb(data, png_ptr, info_ptr);
    
    png_set_IHDR(png_ptr, info_ptr, width, height, 1, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    
    return row_data;
}

png_bytepp convert_bitwise_8bit_grayscale(image_data *data, png_structp png_ptr, png_infop info_ptr) {
    // width as the squareroot of the size times 8 because each bit requires a separate pixel; round app for
    int width = round(sqrt((double)(data->size << 3)));
    // height is size divided by width,
    //int height = round(data->size * 8 / (double)width);
    int height = width + (width * width < data->size << 3);
    // embed image information
    png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    // alloc enough memory to hold the pointers to the individual rows
    png_byte **row_data = calloc(height, sizeof(png_bytep));
    // for every row
    for(int y = 0; y < height; y++) {
        // alloc row memory
        row_data[y] = calloc(width, sizeof(png_byte));
        // for every column
        for(int x = 0; x < width; x++) {
            // get bit index
            int px_i = y * width + x;
            // check if already at EOF/EOB
            if(px_i == (data->size << 3)) return row_data;
            // get on/of state at bit index
            uint8_t value = data->buffer[px_i >> 3] >> (7 - px_i % 8) & 1;
            // store value at the current position in the image data (assigning 266 to uint8 results in 0, as 256 > uint8)
            row_data[y][x] = value * 0xFF;
        }
    }
    
    return row_data;
}

png_bytepp convert_bytewise_8bit_rgb(image_data *data, png_structp png_ptr, png_infop info_ptr) {
    int width = round(sqrt((double)data->size));
    int height = width;
    
    png_bytep *row_data = calloc(height-1, sizeof(png_bytep));
    
    for(int y = 0; y < height - 1; y++) {
        // alloc row memory
        row_data[y] = malloc(width * sizeof(png_byte));
        // literally copy chunks of the source to the PNG
        memcpy(row_data[y], data->buffer + y * width, width);
    }
    // copy the last chunk separatly as it may be smaller than the width -> prevent random data
    row_data[height-1] = calloc(width, sizeof(png_byte));
    printf("Copying last chunk to row %d of length %lu, starting at index %d in buffer\n", height-1, data->size - width*(height-1), (height-1)*width);
    memcpy(row_data[height-1], data->buffer + (height-1)*width, data->size - width*(height-1));
    
    png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    
    return row_data;
}

image_data *bpic_encode_memory(image_data *data, bpic_info* info) {
    int width = 0;
    int height = 0;
    
    if(!data->buffer) {
        throw_err(bpic_error_input_buffer_null, 0);
        return (void*)bpic_error_input_buffer_null;
    }
    if(!info) info = (bpic_info*) malloc(sizeof(bpic_info));
    
    
    if(info->encoding != bpic_bitwise_encoding && info->encoding != bpic_bytewise_encoding) info->encoding = bpic_bitwise_encoding;
    //if(!info->px_size) info->px_size = 4;
    //if(!info->px_size) info->px_size = 1;
    info->px_size = 1;
    
    //if(!data->size) data->size = data->size;
    
    if(info->encoding == bpic_bitwise_encoding) {
        width = round(sqrt((double)(data->size << 3)));
        //width += width % 8;
        height = width;//round(data->size * 8 / (double)width);
    } else if(info->encoding == bpic_bytewise_encoding) {
        width = sqrt((double)data->size);
        height = width + (width * width < data->size);
    } else {
        return NULL;
    }
    
    printf("%u x %u -> %u x %u\n", width, height, width * info->px_size, height * info->px_size);
    
    info->fletcher16 = fletcher16(data->buffer, data->size);
    
    png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    png_infop info_ptr = png_create_info_struct(png_ptr);
    
    //png_bytep image_data = malloc(sizeof(png_byte) * width * height);
    png_byte **row_data = NULL;
    if(!info->encoding) {
        clock_t start_bit8 = clock();
        row_data = convert_bitwise_8bit_grayscale(data, png_ptr, info_ptr); // works
        clock_t end_bit8 = clock();
        printf("8-bit / grayscale; time: %f sec\n", (double)(end_bit8 - start_bit8)/CLOCKS_PER_SEC);
    } else {
        clock_t start_byte = clock();
        row_data = convert_bytewise_8bit_rgb(data, png_ptr, info_ptr); // why u no work
        clock_t end_byte = clock();
        printf("1-byte / grayscale; time: %f sec\n", (double)(end_byte - start_byte)/CLOCKS_PER_SEC);
    }
    
    png_byte **scaled_data = NULL;
    
    // then scale it to its actual size
    if(info->px_size > 1) {
        // cache row size
        int row_size = sizeof(png_byte) * width * info->px_size * 2;
        // create a pointer to the scaled up data
        scaled_data = malloc(height * info->px_size * sizeof(png_bytep));
        // alloc memory here as rows will be allocated that normally would not be in range with the row index
        for(int y = 0; y < height * info->px_size; y++) scaled_data[y] = malloc(row_size);
        // for every [pixelsize]th row
        for(int y = 0; y < height * info->px_size; y+=info->px_size) {
            // for every [pixelsize]th column
            for(int x = 0; x < width * info->px_size; x+=info->px_size) {
                // copy the pixeldata from the 1x1 version to the scaled version; i.e. px_size=4, y=4, x=4 results in row_data[1][1]
                memcpy(&scaled_data[y][x], &row_data[(y / info->px_size)][x / info->px_size], sizeof(png_byte));
                // fill the following pixels with the data at the position in scaled_data, in the previous example y=4 and x=5,6,7
                for(int x_offset = 1; x_offset < info->px_size; x_offset++) memcpy(&scaled_data[y][x+x_offset], &scaled_data[y][x], sizeof(png_byte));
            }
            // copy the previously calculated row "down", meaning the data from y=4 to y=5,6,7
            for(int y_offset = 1; y_offset < info->px_size; y_offset++) memcpy(scaled_data[y+y_offset], scaled_data[y], row_size);
        }
    } else {
        // if the pixel size is actually 1, just point to the already calculated data
        scaled_data = row_data;
    }

    png_bytep bpic_data = calloc(7, sizeof(png_byte));
    
    bpic_data[0] += (uint8_t)sqrt(info->px_size) << 2;
    bpic_data[0] += info->encoding;
    memcpy(bpic_data + 1, &data->size, sizeof(uint32_t));
    memcpy(bpic_data + 5, &info->fletcher16, sizeof(uint16_t));

    for(int byte = 0; byte < 7; byte++) {
        for(int bit = 1; bit < 9; bit++) {
            printf("%s", bpic_data[byte] >> (8 - bit) & 1 ? "1" : "0");
        }
        printf(" ");
    }
    printf("\n");
    
    printf("pxsize: %u; encoding: %u; length: %zu; checksum: %u\n", info->px_size, info->encoding, data->size, info->fletcher16);
    
    // Finally, set up the libpng stuff for writing

    mem_encode *target = malloc(sizeof(mem_encode));
    target->buffer = NULL;
    target->size = 0;
    png_set_write_fn(png_ptr, target, write_to_buffer, NULL);

    png_bytep chunk_data = calloc(7, sizeof(png_byte));
    memcpy(chunk_data, bpic_data, 7);
    
    png_size_t size = 7;
    png_bytep type = (png_bytep)"pvSF";
    
    png_write_info(png_ptr, info_ptr);
    png_write_chunk(png_ptr, type, bpic_data, size);
    png_write_image(png_ptr, scaled_data);
    png_write_end(png_ptr, info_ptr);
    
    //fclose(out);
    
    png_destroy_info_struct(png_ptr, &info_ptr);
    png_destroy_write_struct(&png_ptr, &info_ptr);


    if(info->px_size > 1) {
        for(int y = 0; y < height; y++) free(scaled_data[y]);
        free(scaled_data);
    }
    for(int y = 0; y < height / info->px_size; y++) free(row_data[y]);
    free(row_data);
    free(bpic_data);
    free(chunk_data);
    free(info);

    return target;
}

image_data *bpic_encode_file(FILE *input, bpic_info* info) {
    if(!input) throw_err(bpic_error_input_file_null, 1);
    
    // prepare file
    fseek(input, 0, SEEK_END);
    size_t flen = ftell(input);
    fseek(input, 0, SEEK_SET);
    
    image_data *data = malloc(sizeof(image_data));
    data->buffer = malloc(flen);
    fread(data->buffer, sizeof(unsigned char), flen, input);
    data->size = flen;
    
    encoded_data *output = bpic_encode_memory(data, info);
    
    free(data->buffer);
    free(data);
    
    return output;
}

void write_to_buffer(png_structp png_ptr, png_bytep data, png_size_t length) {
    mem_encode* data_info = png_get_io_ptr(png_ptr);
    size_t size = data_info->size + length;
    
    if(data_info->buffer) {
        data_info->buffer = realloc(data_info->buffer, size);
    } else {
        data_info->buffer = malloc(size);
    }

    memcpy(data_info->buffer + data_info->size, data, length);
    data_info->size += length;
}
