//
//  common_utils.h
//  bpic
//
//  Created by theBreadCompany on 04.03.23.
//

#ifndef common_utils_h
#define common_utils_h

#include <stdbool.h>
#include <png.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <math.h>
#include <time.h>


#pragma mark - Header Information

#define BPIC_BITWISE_ENCODING   0b00
#define BPIC_BYTEWISE_ENCODING  0b01

#define BPIC_UNCOMPRESSED       0b00
#define BPIC_COMPRESSED         0b10

enum bpic_encoding {
    bpic_bitwise_encoding,
    bpic_bytewise_encoding,
    //bpic_bitwise_zipped_encoded,
    //bpic_bytewise_zipped_encoded
};


typedef struct {
    uint8_t px_size;  // first 6 bits (limits px size to up to 2^6-1 = 63, which should be more than sufficient
    enum bpic_encoding encoding; // 7th and 8th bit, allowing for 4 encodings (only two used for now
    size_t length; // 9th - 40 th bit (4 bytes)
    uint16_t fletcher16; // 41st - 56th bit (2 bytes)
    
} bpic_info;

#pragma mark - Helper structs

typedef struct {
    unsigned char *buffer;
    size_t size;
} mem_encode;

typedef mem_encode image_data;
typedef image_data encoded_data;
typedef image_data decoded_data;

typedef struct {
    unsigned char *buffer;
    size_t size;
    png_uint_32 pos;
} mem_decode;

#pragma mark - Crypto

uint16_t fletcher16(const unsigned char *data, size_t size); //

#pragma mark - Error Handling

typedef const char* bpic_message;
extern bpic_message bpic_error_input_file_null;
extern bpic_message bpic_error_input_buffer_null;
extern bpic_message bpic_error_input_buffer_small;
extern bpic_message bpic_warn_encoding_invalid;
extern bpic_message bpic_warn_pxsize_null;


void throw_err(bpic_message message, int error);
void debug(FILE *pipe, const char *message, ...);

#endif /* common_utils_h */
