//
//  decode.h
//  bpic
//
//  Created by theBreadCompany on 04.03.23.
//

#ifndef decode_h
#define decode_h

#include "common_utils.h"


/**
 decode memory
 
 @param data The buffer to decode
 @param checksum_matching A pointer that stores stores whether the checksum matches the one encoded in the image
 @return A pointer to the decoded memory or a bpic\_message on failure.
 */
decoded_data *bpic_decode_memory(encoded_data *data, bool *checksum_matching);
/**
 decode a file
 
 @param input A pointer to a file to decode; NULL is invalid
 @param checksum_matching A pointer that stores stores whether the checksum matches the one encoded in the image
 @returns A pointer to the decoded memory or a bpic\_message on failure.
 */
decoded_data *bpic_decode_file(FILE *input, bool *checksum_matching);

/**
 Decode a png
 
 This method may only be invoced via `bpic_decode_file(FILE *input)` or `bpic_decode_memory(encoded_data *data)`.
 
 @param png_ptr A pointer to the png_ptr that is to be decoded. The input has to be set up, either via `png_set_read_fn` or `png_init_io`
 @param png_info A pointer to a `png_info` struct if already available
 @param checksum_matching A pointer that stores stores whether the checksum matches the one encoded in the image
 @return A pointer to the decoded data
 */
decoded_data *bpic_decode_png(png_structrp png_ptr, png_inforp png_info, bool *checksum_matching);

#endif /* decode_h */
