//
//  main.m
//  bpicencode
//
//  Created by theBreadCompany on 28.02.23.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/time.h>
#include "bpic.h"

// prototype for earlier use
void printHelper(void);
bool processArgv(int argc, const char *argv[], unsigned int *encoding, const char **input, const char **output, unsigned int *bit_depth);

void printHelp(void) {
    printf(
    "Usage: bpicencode <file> [-o <output> -e <encoding> -b <bitdepth>]\n"
    "<file>         the file to encode\n"
    "--output,-o    sets a destination file name; default: <file>.png\n"
    "--encoding,-o  sets a encoding (0 for bitwise, 1 for bytewise); default: 0\n"
    "--bitdepth,-b  sets a bit depth (1 (doesnt work with bytewise) or 8); default: 8\n"
    "file sizes: 8 bit > 1 bit > bytewise encoding; safety is the other way around\n"
    );
}

int main(int argc, const char * argv[]) {
    int encoding;
    FILE *inputFile;
    FILE *outputFile;

    
    if(argc == 1 || strlen(argv[1]) != 1 || !(strcmp(&argv[1][0], "0") || strcmp(&argv[1][0], "1"))) {
        fprintf(stderr, "Encoding has to be either '0' for bitwise or '1' for bytewise!\n");
        return 1;
    }
    encoding = strcmp(argv[1], "0");
    
    inputFile = fopen(argv[2], "rb");
    if(!inputFile) {
        fprintf(stderr, "Could not open file %s!\n", argv[2]);
        return 1;
    }
    
    clock_t start = clock();
    
    bpic_info *info = malloc(sizeof(bpic_info));
    info->encoding = encoding;
    
    decoded_data *data = bpic_encode_file(inputFile, info);
    
    clock_t end = clock();
    printf("Done in %f seconds.\n", (double)(end - start) / CLOCKS_PER_SEC);
    
    uint8_t len = strlen(argv[2]);
    char *oname = malloc(len + 4 + 1); // 4 bytes for ".png" and 1 byte for \0
    strcpy(oname, argv[2]);
    strcpy(oname + len, ".png");
    outputFile = fopen(oname, "wb");
    int code = 0;
    if(fwrite(data->buffer, sizeof(unsigned char), data->size, outputFile) != data->size * sizeof(unsigned char)) {
        fprintf(stderr, "Failed writing output file!");
        code = 1;
    }
    
    fclose(outputFile);
    free(oname);
    free(data->buffer);
    free(data);
    
    return code;
}

bool processArgv(int argc, const char *argv[], unsigned int *encoding, const char **input, const char **output, unsigned int *bit_depth) {
    for(int i = 0; i < argc; i++) {
        if(strcmp(argv[i], "--output") || strcmp(argv[i], "-o")) {
            if(i+1 <= argc || strncmp(argv[i+1], "-", 1)) {
                fprintf(stderr, "missing output value!\n");
                return false;
            } else {
                i++;
                *output = argv[i];
            }
        } else if(strcmp(argv[i], "--bit-depth") || strcmp(argv[i], "-b")) {
            if(i+1 <= argc || strncmp(argv[i+1], "-", 1)) {
                fprintf(stderr, "missing bit depth value!\n");
                return false;
            } else {
                i++;
                *bit_depth = (int)strtol(argv[i], NULL, 10);
            }
        } else if(strcmp(argv[i], "--encoding") || strcmp(argv[i], "-e")) {
            if(i+1 <= argc || strncmp(argv[i+1], "-", 1)) {
                fprintf(stderr, "missing encoding value!\n");
                return false;
            } else {
                i++;
                *encoding = (int)strtol(argv[i], NULL, 10);
            }
        } else if(strncmp(argv[i], "--", 2)) {
            fprintf(stderr, "unknown flag '%s'!\n", argv[i]);
            return false;
        } else {
            FILE *fin = fopen(argv[i], "rb");
            if(fin) {
                *input = argv[i];
                fclose(fin);
            } else {
                fprintf(stderr, "input file does not exist!\n");
                return false;
            }
        }
    }
    
    return true;
}
